#include "scene.h"

/** Construct the scene */
Scene::Scene(const char* name) :
	Model(name),
	// Textures and shaders won't be loaded until .load() is called!
	texture("checkers.png"),
	shader("shader.vert", "shader.frag", NULL),
	spotShader("spotShader.vert", "spotShader.frag", NULL),
	toonShader("toonShader.vert", "toonShader.frag", NULL),
	// Lights
	pointLight("Point Light", GL_LIGHT1,
		/**direction part**/ -5, 5, 5,
		/**diffuse part**/ 1.0f, 0.85f, 0.6f,
		/**specular part**/ 1.0, 0.85f, 0.6f,
		/**ambient part**/ .2f, 0.1f, 0.1f,
		/**attenuation part**/ 0.4, 0.7, 0),
	spotLight("Spot Light", GL_LIGHT2,
		/**direction part**/ 0, 5, 5,
		/**diffuse part**/ 1.0f, 0.85f, 0.6f,
		/**specular part**/ 1.0, 0.85f, 0.6f,
		/**ambient part**/ .2f, 0.1f, 0.1f,
		/**attenuation part**/ 0.4, 0.7, 0,
		/**spot dir part **/ 0,-1,.5, 
		/** cut and falloff **/25, 15 ),
	directionalLight("Directional Light", GL_LIGHT0,
		/**direction part**/ 5, 5, 5,
		/**diffuse part**/ 0.4f, 0.6f, 0.6f,
		/**specular part**/ 0.4f, 0.6f, 0.6f),
	// Sample Model (change this to be your own model)
	sample("My Model"),
	// Properties
	useTexture("Use Texture", true),
	rotateX("Rotate Basic Shape X", -180, 180, 0, 1),
	rotateY("Rotate Basic Shape Y", -180, 180, 0, 1),
	toonLine("Toon Line Width", 0.05, .5, .2, 0.01 ),
	brightness("Brightness", 0.0f, 1.0f, 1.0f, 0.1f),
	diffuse("Diffuse Color", 1.0, 0.7, .4),
	// Shader selector (add your own shaders to here)
	shaderChoice("Shader To Use:", "None|My Shader|Spot Shader|Toon Shader", 0),

	showReferenceUnitSphere("Show Reference Unit Sphere", false),
	// Model selector (add your own model to here instead of Sample Model)
	shapeChoice("Model Shape:", "Sphere|Cube|Cylinder|Torus|Icosahedron|Teapot|Revolution|My Model", 0)

{
	// If you have child Models, like the MobileLight model from model.h,
	// you can add their property groups, and they will appear in the list
	// in the top left corner of Modeler, under this model's entry:
	properties
		.add(pointLight.getProperties())
		.add(directionalLight.getProperties())
		.add(spotLight.getProperties());


	properties.add(sample.getProperties());


	properties.add(&useTexture);

	properties
		.add(&showReferenceUnitSphere)
		.add(&shapeChoice);

	properties			
		.add(&shaderChoice)
		.add(&rotateX)
		.add(&rotateY)
		.add(&brightness)
		.add(&diffuse)
		.add(&toonLine);

}

void Scene::load() {
	// Add any textures and shaders you want to load into here,
	// this is called whenever the Reload button is pressed.
	texture.load();
	shader.load();
	spotShader.load();
	toonShader.load();
}

void Scene::draw() {
	// The lights must be drawn FIRST, so the other scene elements
	// can get lit!
	pointLight.draw();
	directionalLight.draw();
	spotLight.draw();
	//reset colors
	setDiffuseColor(
		diffuse.getRed(),
		diffuse.getGreen(),
		diffuse.getBlue());

	setAmbientColor(0.2, 0.2, 0.2);

	setSpecularColor(1, 1, 1);
	setShininess(pow(2.71828183f, 3));

	// Use the texture if desired.
	if(useTexture.getValue()) {
		texture.use();
	} else {
		glBindTexture(GL_TEXTURE_2D, 0);
	}

	GLint brightnessVariableLocation;
	GLfloat lineWidthLocation;
	switch(shaderChoice.getValue()) {
		case 0:
			//glBindTexture(GL_TEXTURE_2D, 0); // don't use texture
			if(glUseProgram) { glUseProgram(0); } // don't use shader
			break;
		case 1: // My Shader.
			shader.use();
			// glGetUniformLocation gets the memory location of a variable with the given char* name, in this case "brightness"
			// for the given shader program, identified by its ID.
			brightnessVariableLocation = glGetUniformLocation(shader.getID(), "brightness");
			// glUniform1f sets the value of a particular uniform variable location with a single float value (hence the suffix "1f")
			glUniform1f(brightnessVariableLocation, brightness.getValue());
			break;
		case 2: //Spot Shader.
			spotShader.use();
			brightnessVariableLocation = glGetUniformLocation(spotShader.getID(), "brightness");
			glUniform1f(brightnessVariableLocation, brightness.getValue());
			break;
		case 3: //Toon Shader.
			toonShader.use();
			lineWidthLocation = glGetUniformLocation(toonShader.getID(), "lineWidth");
			glUniform1f(lineWidthLocation, toonLine.getValue());
			break;
	}


	// Call a class method that draws our model.
	drawModel();
	// Stop applying shaders to objects.
	// This if-statement makes sure that glUseProgram is not a null
	// function pointer (which it will be if GLEW couldn't initialize).
	if(glUseProgram) {
		glUseProgram(0);
	}

	// Stop applying textures to objects
	glBindTexture(GL_TEXTURE_2D, 0);
}

void Scene::drawModel() {
	// Level of Detail
	int divisions = 16;
	switch(ModelerDrawState::Instance()->m_quality) {
		case HIGH:
			divisions = 64; break;
		case MEDIUM:
			divisions = 20; break;
		case LOW:
			divisions = 12; break;
		case POOR:
			divisions = 8; break;
	}

	// Remember our current modelview matrix, which is in world space.
	glPushMatrix();

	// Rotate around the X-axis
	glRotatef(rotateX.getValue(), 1, 0, 0);

	// Rotate around the Y-axis
	glRotatef(rotateY.getValue(), 0, 1, 0);


	// Draw some object at the origin.
	switch(shapeChoice.getValue()) {
		case 0: // sphere
			glRotatef(-90.f, 1.f, 0.f, 0.f);
			drawSphere(1.0);
			// glutSolidSphere(1.0, divisions, divisions);
			glRotatef(90.f, 1.f, 0.f, 0.f);
			break;
		case 1: // box
			glRotatef(-45.f, 0.f, 1.f, 0.f);
			glutSolidCube(2);
			glRotatef(45.f, 0.f, 1.f, 0.f);
			break;
		case 2: // cylinder
			glRotatef(90.f, 1.f, 0.f, 0.f);
			drawCylinder(1.8, 1.0, 1.8);
			glRotatef(-90.f, 1.f, 0.f, 0.f);
			break;
		case 3: // torus
			glutSolidTorus(.5, 1, divisions, divisions);
			break;
		case 4: // icosahedron
			glutSolidIcosahedron();
			break;
		case 5: // teapot
			glutSolidTeapot(1);
			break;
		case 6: // surface of revolution
			drawRevolution(1);
			break;
		case 7: // Replaced with My Model
			sample.draw();
			break;
	}

	if(showReferenceUnitSphere.getValue()) {
		// Now go 1 units in the +y direction,
		// and 1 units in the NEGATIVE x direction.
		glTranslated(-2, -1, 0);
		// Now go 1 units in the +y direction,
		// and 1 units in the NEGATIVE x direction.
		glRotatef(-90.f, 1.f, 0.f, 0.f);
		drawSphere(.5);
		glRotatef(90.f, 1.f, 0.f, 0.f);
		glPopMatrix();
	}

	// Go back to "world space"
	glPopMatrix();

	{

		// Draw the floor.  Here's a basic example of texture mapping.
		glBegin(GL_QUADS);
			glTexCoord2f(0, 1); // specify the texture coordinate
			glNormal3f(0, 1, 0); // specify the surface's normal at this vertex
			glVertex3f(-10, -2, -10); // both before its corresponding vertex

			glTexCoord2f(1, 1);
			glNormal3f(0, 1, 0);
			glVertex3f(10, -2, -10);

			glTexCoord2f(1, 0);
			glNormal3f(0, 1, 0);
			glVertex3f(10, -2, 10);

			glTexCoord2f(0, 0);
			glNormal3f(0, 1, 0);
			glVertex3f(-10, -2, 10);
		glEnd();
	}
}

void Scene::tick() {
	// You can put code here to increment animation counters for
	// extra credit.
}

void Scene::swapTexture(const char* filename) {
	texture = Texture2D(filename);
	texture.load();
}



